# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Library to assist processing Snapshot Metadata protos into text"""

from typing import Optional, List, Mapping
import pw_log_tokenized
import pw_tokenizer
from pw_tokenizer import proto as proto_detokenizer
from pw_snapshot_metadata_proto import snapshot_metadata_pb2

_PRETTY_FORMAT_DEFAULT_WIDTH = 80

_FATAL = (
    '▪▄▄▄ ▄▄▄· ▄▄▄▄▄ ▄▄▄· ▄ ·',
    '█▄▄▄▐█ ▀█ • █▌ ▐█ ▀█ █  ',
    '█ ▪ ▄█▀▀█   █. ▄█▀▀█ █  ',
    '▐▌ .▐█ ▪▐▌ ▪▐▌·▐█ ▪▐▌▐▌ ',
    '▀    ▀  ▀ ·  ▀  ▀  ▀ .▀▀',
)


def _process_tags(tags: Mapping[str, str]) -> Optional[str]:
    """Outputs snapshot tags as a multi-line string."""
    if not tags:
        return None

    output: List[str] = ['Tags:']
    for key, value in tags.items():
        output.append(f'  {key}: {value}')

    return '\n'.join(output)


def process_snapshot(
    serialized_snapshot: bytes, tokenizer_db: Optional[pw_tokenizer.Detokenizer]
) -> str:
    """Processes snapshot metadata and tags, producing a multi-line string."""
    snapshot = snapshot_metadata_pb2.SnapshotBasicInfo()
    snapshot.ParseFromString(serialized_snapshot)

    output: List[str] = []

    if snapshot.HasField('metadata'):
        output.extend(
            (
                str(MetadataProcessor(snapshot.metadata, tokenizer_db)),
                '',
            )
        )

    if snapshot.tags:
        tags = _process_tags(snapshot.tags)
        if tags:
            output.append(tags)
        # Trailing blank line for spacing.
        output.append('')

    return '\n'.join(output)


class MetadataProcessor:
    """This class simplifies dumping contents of a snapshot Metadata message."""

    def __init__(
        self,
        metadata: snapshot_metadata_pb2.Metadata,
        tokenizer_db: Optional[pw_tokenizer.Detokenizer],
    ):
        self._metadata = metadata
        self._tokenizer_db = (
            tokenizer_db
            if tokenizer_db is not None
            else pw_tokenizer.Detokenizer(None)
        )
        self._reason_token = self._tokenizer_db.detokenize(
            metadata.reason
        ).token
        self._format_width = _PRETTY_FORMAT_DEFAULT_WIDTH
        proto_detokenizer.detokenize_fields(self._tokenizer_db, self._metadata)

    def is_fatal(self) -> bool:
        return self._metadata.fatal

    def reason(self) -> str:
        if not self._metadata.reason:
            return 'UNKNOWN (field missing)'

        log = pw_log_tokenized.FormatStringWithMetadata(
            self._metadata.reason.decode()
        )

        return f'{log.file}: {log.message}' if log.file else log.message

    def reason_token(self) -> Optional[int]:
        """If the snapshot `reason` is tokenized, the value of the token."""
        return self._reason_token

    def project_name(self) -> str:
        return self._metadata.project_name.decode()

    def device_name(self) -> str:
        return self._metadata.device_name.decode()

    def device_fw_version(self) -> str:
        return self._metadata.software_version

    def snapshot_uuid(self) -> str:
        return self._metadata.snapshot_uuid.hex()

    def fw_build_uuid(self) -> str:
        return self._metadata.software_build_uuid.hex()

    def set_pretty_format_width(self, width: int):
        """Sets the centered width of the FATAL text for a formatted output."""
        self._format_width = width

    def __str__(self) -> str:
        """outputs a pw.snapshot.Metadata proto as a multi-line string."""
        output: List[str] = []
        if self._metadata.fatal:
            output.extend(
                (
                    *[x.center(self._format_width).rstrip() for x in _FATAL],
                    '',
                    'Device crash cause:',
                )
            )
        else:
            output.append('Snapshot capture reason:')

        output.extend(
            (
                '    ' + self.reason(),
                '',
            )
        )
        if self.reason_token():
            output.append(f'Reason token:      0x{self.reason_token():x}')

        if self._metadata.project_name:
            output.append(f'Project name:      {self.project_name()}')

        if self._metadata.device_name:
            output.append(f'Device:            {self.device_name()}')

        if self._metadata.software_version:
            output.append(f'Device FW version: {self.device_fw_version()}')

        if self._metadata.software_build_uuid:
            output.append(f'FW build UUID:     {self.fw_build_uuid()}')

        if self._metadata.snapshot_uuid:
            output.append(f'Snapshot UUID:     {self.snapshot_uuid()}')

        return '\n'.join(output)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
"""OWNERS file checks."""

import argparse
import collections
import dataclasses
import difflib
import enum
import functools
import logging
import pathlib
import re
import sys
from typing import (
    Callable,
    Collection,
    DefaultDict,
    Dict,
    Iterable,
    List,
    OrderedDict,
    Set,
    Union,
)
from pw_presubmit import git_repo
from pw_presubmit.presubmit import PresubmitFailure

_LOG = logging.getLogger(__name__)


class LineType(enum.Enum):
    COMMENT = "comment"
    WILDCARD = "wildcard"
    FILE_LEVEL = "file_level"
    FILE_RULE = "file_rule"
    INCLUDE = "include"
    PER_FILE = "per-file"
    USER = "user"
    # Special type to hold lines that don't get attached to another type
    TRAILING_COMMENTS = "trailing-comments"


_LINE_TYPERS: OrderedDict[
    LineType, Callable[[str], bool]
] = collections.OrderedDict(
    (
        (LineType.COMMENT, lambda x: x.startswith("#")),
        (LineType.WILDCARD, lambda x: x == "*"),
        (LineType.FILE_LEVEL, lambda x: x.startswith("set ")),
        (LineType.FILE_RULE, lambda x: x.startswith("file:")),
        (LineType.INCLUDE, lambda x: x.startswith("include ")),
        (LineType.PER_FILE, lambda x: x.startswith("per-file ")),
        (
            LineType.USER,
            lambda x: bool(re.match("^[a-zA-Z1-9.+-]+@[a-zA-Z0-9.-]+", x)),
        ),
    )
)


class OwnersError(Exception):
    """Generic level OWNERS file error."""

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message


class FormatterError(OwnersError):
    """Errors where formatter doesn't know how to act."""


class OwnersDuplicateError(OwnersError):
    """Errors where duplicate lines are found in OWNERS files."""


class OwnersUserGrantError(OwnersError):
    """Invalid user grant, * is used with any other grant."""


class OwnersProhibitedError(OwnersError):
    """Any line that is prohibited by the owners syntax.

    https://android-review.googlesource.com/plugins/code-owners/Documentation/backend-find-owners.html
    """


class OwnersDependencyError(OwnersError):
    """OWNERS file tried to import file that does not exists."""


class OwnersInvalidLineError(OwnersError):
    """Line in OWNERS file does not match any 'line_typer'."""


class OwnersStyleError(OwnersError):
    """OWNERS file does not match style guide."""


@dataclasses.dataclass
class Line:
    content: str
    comments: List[str] = dataclasses.field(default_factory=list)


class OwnersFile:
    """Holds OWNERS file in easy to use parsed structure."""

    path: pathlib.Path
    original_lines: List[str]
    sections: Dict[LineType, List[Line]]
    formatted_lines: List[str]

    def __init__(self, path: pathlib.Path) -> None:
        if not path.exists():
            error_msg = f"Tried to import {path} but it does not exists"
            raise OwnersDependencyError(error_msg)
        self.path = path

        self.original_lines = self.load_owners_file(self.path)
        cleaned_lines = self.clean_lines(self.original_lines)
        self.sections = self.parse_owners(cleaned_lines)
        self.formatted_lines = self.format_sections(self.sections)

    @staticmethod
    def load_owners_file(owners_file: pathlib.Path) -> List[str]:
        return owners_file.read_text().split("\n")

    @staticmethod
    def clean_lines(dirty_lines: List[str]) -> List[str]:
        """Removes extra whitespace from list of strings."""

        cleaned_lines = []
        for line in dirty_lines:
            line = line.strip()  # Remove initial and trailing whitespace

            # Compress duplicated whitespace and remove tabs.
            # Allow comment lines to break this rule as they may have initial
            # whitespace for lining up text with neighboring lines.
            if not line.startswith("#"):
                line = re.sub(r"\s+", " ", line)
            if line:
                cleaned_lines.append(line)
        return cleaned_lines

    @staticmethod
    def __find_line_type(line: str) -> LineType:
        for line_type, type_matcher in _LINE_TYPERS.items():
            if type_matcher(line):
                return line_type

        raise OwnersInvalidLineError(
            f"Unrecognized OWNERS file line, '{line}'."
        )

    @staticmethod
    def parse_owners(
        cleaned_lines: List[str],
    ) -> DefaultDict[LineType, List[Line]]:
        """Converts text lines of OWNERS into structured object."""
        sections: DefaultDict[LineType, List[Line]] = collections.defaultdict(
            list
        )
        comment_buffer: List[str] = []

        def add_line_to_sections(
            sections, section: LineType, line: str, comment_buffer: List[str]
        ):
            if any(
                seen_line.content == line for seen_line in sections[section]
            ):
                raise OwnersDuplicateError(f"Duplicate line '{line}'.")
            line_obj = Line(content=line, comments=comment_buffer)
            sections[section].append(line_obj)

        for line in cleaned_lines:
            line_type: LineType = OwnersFile.__find_line_type(line)
            if line_type == LineType.COMMENT:
                comment_buffer.append(line)
            else:
                add_line_to_sections(sections, line_type, line, comment_buffer)
                comment_buffer = []

        add_line_to_sections(
            sections, LineType.TRAILING_COMMENTS, "", comment_buffer
        )

        return sections

    @staticmethod
    def format_sections(
        sections: DefaultDict[LineType, List[Line]]
    ) -> List[str]:
        """Returns ideally styled OWNERS file.

        The styling rules are
        * Content will be sorted in the following orders with a blank line
        separating
            * "set noparent"
            * "include" lines
            * "file:" lines
            * user grants (example, "*", foo@example.com)
            * "per-file:" lines
        * Do not combine user grants and "*"
        * User grants should be sorted alphabetically (this assumes English
        ordering)

        Returns:
          List of strings that make up a styled version of a OWNERS file.

        Raises:
          FormatterError: When formatter does not handle all lines of input.
                          This is a coding error in owners_checks.
        """
        all_sections = [
            LineType.FILE_LEVEL,
            LineType.INCLUDE,
            LineType.FILE_RULE,
            LineType.WILDCARD,
            LineType.USER,
            LineType.PER_FILE,
            LineType.TRAILING_COMMENTS,
        ]
        formatted_lines: List[str] = []

        def append_section(line_type):
            # Add a line of separation if there was a previous section and our
            # current section has any content. I.e. do not lead with padding and
            # do not have multiple successive lines of padding.
            if (
                formatted_lines
                and line_type != LineType.TRAILING_COMMENTS
                and sections[line_type]
            ):
                formatted_lines.append("")

            sections[line_type].sort(key=lambda line: line.content)
            for line in sections[line_type]:
                # Strip keep-sorted comments out since sorting is done by this
                # script
                formatted_lines.extend(
                    [
                        comment
                        for comment in line.comments
                        if not comment.startswith("# keep-sorted: ")
                    ]
                )
                formatted_lines.append(line.content)

        for section in all_sections:
            append_section(section)

        if any(section_name not in all_sections for section_name in sections):
            raise FormatterError("Formatter did not process all sections.")
        return formatted_lines

    def check_style(self) -> None:
        """Checks styling of OWNERS file.

        Enforce consistent style on OWNERS file. This also incidentally detects
        a few classes of errors.

        Raises:
          OwnersStyleError: Indicates styled lines do not match original input.
        """

        if self.original_lines != self.formatted_lines:
            print(
                "\n".join(
                    difflib.unified_diff(
                        self.original_lines,
                        self.formatted_lines,
                        fromfile=str(self.path),
                        tofile="styled",
                        lineterm="",
                    )
                )
            )

            raise OwnersStyleError(
                "OWNERS file format does not follow styling."
            )

    def look_for_owners_errors(self) -> None:
        """Scans owners files for invalid or useless content."""

        # Confirm when using the wildcard("*") we don't also try to use
        # individual user grants.
        if self.sections[LineType.WILDCARD] and self.sections[LineType.USER]:
            raise OwnersUserGrantError(
                "Do not use '*' with individual user "
                "grants, * already applies to all users."
            )

        # NOTE: Using the include keyword in combination with a per-file rule is
        # not possible.
        # https://android-review.googlesource.com/plugins/code-owners/Documentation/backend-find-owners.html#syntax:~:text=NOTE%3A%20Using%20the%20include%20keyword%20in%20combination%20with%20a%20per%2Dfile%20rule%20is%20not%20possible.
        if self.sections[LineType.INCLUDE] and self.sections[LineType.PER_FILE]:
            raise OwnersProhibitedError(
                "'include' cannot be used with 'per-file'."
            )

    def __complete_path(self, sub_owners_file_path) -> pathlib.Path:
        """Always return absolute path."""
        # Absolute paths start with the git/project root
        if sub_owners_file_path.startswith("/"):
            root = git_repo.root(self.path)
            full_path = root / sub_owners_file_path[1:]
        else:
            # Relative paths start with owners file dir
            full_path = self.path.parent / sub_owners_file_path
        return full_path.resolve()

    def get_dependencies(self) -> List[pathlib.Path]:
        """Finds owners files this file includes."""
        dependencies = []
        # All the includes
        for include in self.sections.get(LineType.INCLUDE, []):
            file_str = include.content[len("include ") :]
            dependencies.append(self.__complete_path(file_str))

        # all file: rules:
        for file_rule in self.sections.get(LineType.FILE_RULE, []):
            file_str = file_rule.content[len("file:") :]
            if ":" in file_str:
                _LOG.warning(
                    "TODO(b/254322931): This check does not yet support "
                    "<project> or <branch> in a file: rule"
                )
                _LOG.warning(
                    "It will not check line '%s' found in %s",
                    file_rule.content,
                    self.path,
                )

            dependencies.append(self.__complete_path(file_str))

        # all the per-file rule includes
        for per_file in self.sections.get(LineType.PER_FILE, []):
            file_str = per_file.content[len("per-file ") :]
            access_grant = file_str[file_str.index("=") + 1 :]
            if access_grant.startswith("file:"):
                dependencies.append(
                    self.__complete_path(access_grant[len("file:") :])
                )

        return dependencies

    def write_formatted(self) -> None:
        self.path.write_text("\n".join(self.formatted_lines))


def resolve_owners_tree(root_owners: pathlib.Path) -> List[OwnersFile]:
    """Given a starting OWNERS file return it and all of it's dependencies."""
    found = []
    todo = collections.deque((root_owners,))
    checked: Set[pathlib.Path] = set()
    while todo:
        cur_file = todo.popleft()
        checked.add(cur_file)
        owners_obj = OwnersFile(cur_file)
        found.append(owners_obj)
        new_dependents = owners_obj.get_dependencies()
        for new_dep in new_dependents:
            if new_dep not in checked and new_dep not in todo:
                todo.append(new_dep)
    return found


def _run_owners_checks(owners_obj: OwnersFile) -> None:
    owners_obj.look_for_owners_errors()
    owners_obj.check_style()


def _format_owners_file(owners_obj: OwnersFile) -> None:
    owners_obj.look_for_owners_errors()

    if owners_obj.original_lines != owners_obj.formatted_lines:
        owners_obj.write_formatted()


def _list_unwrapper(
    func, list_or_path: Union[Iterable[pathlib.Path], pathlib.Path]
) -> Dict[pathlib.Path, str]:
    """Decorator that accepts Paths or list of Paths and iterates as needed."""
    errors: Dict[pathlib.Path, str] = {}
    if isinstance(list_or_path, Iterable):
        files = list_or_path
    else:
        files = (list_or_path,)

    all_owners_obj: List[OwnersFile] = []
    for file in files:
        all_owners_obj.extend(resolve_owners_tree(file))

    checked: Set[pathlib.Path] = set()
    for current_owners in all_owners_obj:
        # Ensure we don't check the same file twice
        if current_owners.path in checked:
            continue
        checked.add(current_owners.path)
        try:
            func(current_owners)
        except OwnersError as err:
            errors[current_owners.path] = err.message
            _LOG.error(
                "%s: %s", str(current_owners.path.absolute()), err.message
            )
    return errors


# This generates decorated versions of the functions that can used with both
# formatter (which supplies files one at a time) and presubmits (which supplies
# list of files).
run_owners_checks = functools.partial(_list_unwrapper, _run_owners_checks)
format_owners_file = functools.partial(_list_unwrapper, _format_owners_file)


def presubmit_check(
    files: Union[pathlib.Path, Collection[pathlib.Path]]
) -> None:
    errors = run_owners_checks(files)
    if errors:
        for file in errors:
            _LOG.warning("  pw format --fix %s", file)
        _LOG.warning("will automatically fix this.")
        raise PresubmitFailure


def main() -> int:
    """Standalone test of styling."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--style", action="store_true")
    parser.add_argument("--owners_file", required=True, type=str)
    args = parser.parse_args()

    try:
        owners_obj = OwnersFile(pathlib.Path(args.owners_file))
        owners_obj.look_for_owners_errors()
        owners_obj.check_style()
    except OwnersError as err:
        _LOG.error("%s %s", err, err.message)
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Argument parsing code for presubmit checks."""

import argparse
import logging
import os
from pathlib import Path
import re
import shutil
import textwrap
from typing import Callable, Collection, List, Optional, Sequence

from pw_presubmit import git_repo, presubmit

_LOG = logging.getLogger(__name__)
DEFAULT_PATH = Path('out', 'presubmit')

_OUTPUT_PATH_README = '''\
This directory was created by pw_presubmit to run presubmit checks for the
{repo} repository. This directory is not used by the regular GN or CMake Ninja
builds. It may be deleted safely.
'''


def add_path_arguments(parser) -> None:
    """Adds common presubmit check options to an argument parser."""

    parser.add_argument(
        'paths',
        metavar='pathspec',
        nargs='*',
        help=(
            'Paths or patterns to which to restrict the checks. These are '
            'interpreted as Git pathspecs. If --base is provided, only '
            'paths changed since that commit are checked.'
        ),
    )

    base = parser.add_mutually_exclusive_group()
    base.add_argument(
        '-b',
        '--base',
        metavar='commit',
        default=git_repo.TRACKING_BRANCH_ALIAS,
        help=(
            'Git revision against which to diff for changed files. '
            'Default is the tracking branch of the current branch.'
        ),
    )

    base.add_argument(
        '--all',
        '--full',
        dest='base',
        action='store_const',
        const=None,
        help='Run actions for all files, not just changed files.',
    )

    parser.add_argument(
        '-e',
        '--exclude',
        metavar='regular_expression',
        default=[],
        action='append',
        type=re.compile,
        help=(
            'Exclude paths matching any of these regular expressions, '
            "which are interpreted relative to each Git repository's root."
        ),
    )


def _add_programs_arguments(
    parser: argparse.ArgumentParser, programs: presubmit.Programs, default: str
):
    def presubmit_program(arg: str) -> presubmit.Program:
        if arg not in programs:
            all_program_names = ', '.join(sorted(programs.keys()))
            raise argparse.ArgumentTypeError(
                f'{arg} is not the name of a presubmit program\n\n'
                f'Valid Programs:\n{all_program_names}'
            )

        return programs[arg]

    # This argument is used to copy the default program into the argparse
    # namespace argument. It's not intended to be set by users.
    parser.add_argument(
        '--default-program',
        default=[presubmit_program(default)],
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        '-p',
        '--program',
        choices=programs.values(),
        type=presubmit_program,
        action='append',
        default=[],
        help='Which presubmit program to run',
    )

    parser.add_argument(
        '--list-steps-file',
        dest='list_steps_file',
        type=Path,
        help=argparse.SUPPRESS,
    )

    all_steps = programs.all_steps()

    def list_steps() -> None:
        """List all available presubmit steps and their docstrings."""
        for step in sorted(all_steps.values(), key=str):
            _LOG.info('%s', step)
            if step.doc:
                first, *rest = step.doc.split('\n', 1)
                _LOG.info('  %s', first)
                if rest and _LOG.isEnabledFor(logging.DEBUG):
                    for line in textwrap.dedent(*rest).splitlines():
                        _LOG.debug('  %s', line)

    parser.add_argument(
        '--list-steps',
        action='store_const',
        const=list_steps,
        default=None,
        help='List all the available steps.',
    )

    def presubmit_step(arg: str) -> presubmit.Check:
        if arg not in all_steps:
            all_step_names = ', '.join(sorted(all_steps.keys()))
            raise argparse.ArgumentTypeError(
                f'{arg} is not the name of a presubmit step\n\n'
                f'Valid Steps:\n{all_step_names}'
            )
        return all_steps[arg]

    parser.add_argument(
        '--step',
        action='append',
        choices=all_steps.values(),
        default=[],
        help='Run specific steps instead of running a full program.',
        type=presubmit_step,
    )

    parser.add_argument(
        '--substep',
        action='store',
        help=(
            "Run a specific substep of a step. Only supported if there's only "
            'one --step argument and no --program arguments.'
        ),
    )

    def gn_arg(argument):
        key, value = argument.split('=', 1)
        return (key, value)

    # Recipe code for handling builds with pre-release toolchains requires the
    # ability to pass through GN args. This ability is not expected to be used
    # directly outside of this case, so the option is hidden. Values passed in
    # to this argument should be of the form 'key=value'.
    parser.add_argument(
        '--override-gn-arg',
        dest='override_gn_args',
        action='append',
        type=gn_arg,
        help=argparse.SUPPRESS,
    )


def add_arguments(
    parser: argparse.ArgumentParser,
    programs: Optional[presubmit.Programs] = None,
    default: str = '',
) -> None:
    """Adds common presubmit check options to an argument parser."""

    add_path_arguments(parser)
    parser.add_argument(
        '-k',
        '--keep-going',
        action='store_true',
        help='Continue running presubmit steps after a failure.',
    )
    parser.add_argument(
        '--continue-after-build-error',
        action='store_true',
        help=(
            'Within presubmit steps, continue running build steps after a '
            'failure.'
        ),
    )
    parser.add_argument(
        '--output-directory',
        type=Path,
        help=f'Output directory (default: {"<repo root>" / DEFAULT_PATH})',
    )
    parser.add_argument(
        '--package-root',
        type=Path,
        help='Package root directory (default: <env directory>/packages)',
    )

    exclusive = parser.add_mutually_exclusive_group()
    exclusive.add_argument(
        '--clear',
        '--clean',
        action='store_true',
        help='Delete the presubmit output directory and exit.',
    )

    if programs:
        if not default:
            raise ValueError('A default must be provided with programs')

        _add_programs_arguments(parser, programs, default)

        # LUCI builders extract the list of steps from the program and run them
        # individually for a better UX in MILO.
        parser.add_argument(
            '--only-list-steps',
            action='store_true',
            help=argparse.SUPPRESS,
        )


def run(  # pylint: disable=too-many-arguments
    default_program: Optional[presubmit.Program],
    program: Sequence[presubmit.Program],
    step: Sequence[presubmit.Check],
    substep: str,
    output_directory: Optional[Path],
    package_root: Path,
    clear: bool,
    root: Optional[Path] = None,
    repositories: Collection[Path] = (),
    only_list_steps=False,
    list_steps: Optional[Callable[[], None]] = None,
    **other_args,
) -> int:
    """Processes arguments from add_arguments and runs the presubmit.

    Args:
      default_program: program to use if neither --program nor --step is used
      program: from the --program option
      step: from the --step option
      substep: from the --substep option
      output_directory: from --output-directory option
      package_root: from --package-root option
      clear: from the --clear option
      root: base path from which to run presubmit checks; defaults to the root
          of the current directory's repository
      repositories: roots of Git repositories on which to run presubmit checks;
          defaults to the root of the current directory's repository
      only_list_steps: list the steps that would be executed, one per line,
          instead of executing them
      list_steps: list the steps that would be executed with their docstrings
      **other_args: remaining arguments defined by by add_arguments

    Returns:
      exit code for sys.exit; 0 if successful, 1 if an error occurred
    """
    if root is None:
        root = git_repo.root()

    if not repositories:
        repositories = [root]

    if output_directory is None:
        output_directory = root / DEFAULT_PATH

    output_directory.mkdir(parents=True, exist_ok=True)
    output_directory.joinpath('README.txt').write_text(
        _OUTPUT_PATH_README.format(repo=root)
    )

    if not package_root:
        package_root = Path(os.environ['PW_PACKAGE_ROOT'])

    _LOG.debug('Using environment at %s', output_directory)

    if clear:
        _LOG.info('Clearing presubmit output directory')

        if output_directory.exists():
            shutil.rmtree(output_directory)
            _LOG.info('Deleted %s', output_directory)

        return 0

    if list_steps:
        list_steps()
        return 0

    final_program: Optional[presubmit.Program] = None
    if not program and not step:
        assert default_program  # Cast away Optional[].
        final_program = default_program
    elif len(program) == 1 and not step:
        final_program = program[0]
    else:
        steps: List[presubmit.Check] = []
        steps.extend(step)
        for prog in program:
            steps.extend(prog)
        final_program = presubmit.Program('', steps)

    if substep and len(final_program) > 1:
        _LOG.error('--substep not supported if there are multiple steps')
        return 1

    if presubmit.run(
        final_program,
        root,
        repositories,
        only_list_steps=only_list_steps,
        output_directory=output_directory,
        package_root=package_root,
        substep=substep,
        **other_args,
    ):
        return 0

    return 1



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Transforms a JSON list of paths using -ffile-prefix-map style rules."""

import argparse
import json
from typing import Iterator, List, TextIO

# Note: This should be List[Tuple[str, str]], but using string.split()
# produces Tuple[Any,...], so this permits that typing for convenience.
PrefixMaps = List[tuple]


def _parse_args() -> argparse.Namespace:
    """Parses and returns the command line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'in_json',
        type=argparse.FileType('r'),
        help='The JSON file containing a list of file names '
        'that the prefix map operations should be applied to',
    )
    parser.add_argument(
        '--prefix-map-json',
        type=argparse.FileType('r'),
        required=True,
        help=(
            'JSON file containing an array of prefix map transformations to '
            'apply to the strings before tokenizing. These string literal '
            'transformations are of the form "from=to". All strings with the '
            'prefix `from` will have the prefix replaced with `to`. '
            'Transformations are applied in the order they are listed in the '
            'JSON file.'
        ),
    )

    parser.add_argument(
        '--output',
        type=argparse.FileType('w'),
        help='File path to write transformed paths to.',
    )
    return parser.parse_args()


def remap_paths(paths: List[str], prefix_maps: PrefixMaps) -> Iterator[str]:
    for path in paths:
        for from_prefix, to_prefix in prefix_maps:
            if path.startswith(from_prefix):
                path = path.replace(from_prefix, to_prefix, 1)
                break  # Only the first -ffile-prefix-map option applies.
        yield path


def remap_json_paths(
    in_json: TextIO, output: TextIO, prefix_map_json: TextIO
) -> None:
    paths = json.load(in_json)
    prefix_maps: PrefixMaps = [
        tuple(m.split('=', maxsplit=1)) for m in json.load(prefix_map_json)
    ]

    json.dump(list(remap_paths(paths, prefix_maps)), output)


if __name__ == '__main__':
    remap_json_paths(**vars(_parse_args()))



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""The Pigweed command line interface."""

import logging
import os
import sys
from typing import NoReturn

import pw_cli.log
from pw_cli import arguments, plugins, pw_command_plugins

_LOG = logging.getLogger(__name__)


def main() -> NoReturn:
    """Entry point for the pw command."""

    args = arguments.parse_args()

    pw_cli.log.install(level=args.loglevel, debug_log=args.debug_log)

    # Start with the most critical part of the Pigweed command line tool.
    if not args.no_banner:
        arguments.print_banner()

    _LOG.debug('Executing the pw command from %s', args.directory)
    os.chdir(args.directory)

    pw_command_plugins.register(args.directory)

    if args.help or args.command is None:
        print(pw_command_plugins.format_help(), file=sys.stderr)
        sys.exit(0)

    try:
        sys.exit(pw_command_plugins.run(args.command, args.plugin_args))
    except (plugins.Error, KeyError) as err:
        _LOG.critical('Cannot run command %s.', args.command)
        _LOG.critical('%s', err)
        sys.exit(2)


if __name__ == '__main__':
    main()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""A channel adapter layer that introduces lossy behavior to a channel."""

import abc
import collections
import copy
import logging
from typing import Callable, Deque, Optional, Tuple
import random
import time

import pw_rpc

_LOG = logging.getLogger(__name__)


# TODO(amontanez): The surface are of this API could be reduced significantly
# with a few changes to LossyChannel.
class LossController(abc.ABC):
    """Interface for driving loss/corruption decisions of a LossyChannel."""

    @abc.abstractmethod
    def next_packet_duplicated(self) -> bool:
        """Returns true if the next packet should be duplicated."""

    @abc.abstractmethod
    def next_packet_out_of_order(self) -> bool:
        """Returns true if the next packet should be a re-ordered packet."""

    @abc.abstractmethod
    def next_packet_delayed(self) -> bool:
        """Returns true if a delay should occur before sending a packet."""

    @abc.abstractmethod
    def next_packet_dropped(self) -> bool:
        """Returns true if the next incoming packet should be dropped."""

    @abc.abstractmethod
    def next_packet_delay(self) -> int:
        """The delay before sending the next packet, in milliseconds."""

    @abc.abstractmethod
    def next_num_dupes(self) -> int:
        """Returns how many times the next packet should be duplicated."""

    @abc.abstractmethod
    def choose_out_of_order_packet(self, max_idx) -> int:
        """Returns the index of the next reordered packet.

        A return value of 0 represents the newest packet, while max_idx
        represents the oldest packet.
        """


class ManualPacketFilter(LossController):
    """Determines if a packet should be kept or dropped for testing purposes."""

    _Action = Callable[[int], Tuple[bool, bool]]
    _KEEP = lambda _: (True, False)
    _DROP = lambda _: (False, False)

    def __init__(self) -> None:
        self.packet_count = 0
        self._actions: Deque[ManualPacketFilter._Action] = collections.deque()

    def reset(self) -> None:
        self.packet_count = 0
        self._actions.clear()

    def keep(self, count: int) -> None:
        """Keeps the next count packets."""
        self._actions.extend(ManualPacketFilter._KEEP for _ in range(count))

    def drop(self, count: int) -> None:
        """Drops the next count packets."""
        self._actions.extend(ManualPacketFilter._DROP for _ in range(count))

    def drop_every(self, every: int) -> None:
        """Drops every Nth packet forever."""
        self._actions.append(lambda count: (count % every != 0, True))

    def randomly_drop(self, one_in: int, gen: random.Random) -> None:
        """Drops packets randomly forever."""
        self._actions.append(lambda _: (gen.randrange(one_in) != 0, True))

    def keep_packet(self) -> bool:
        """Returns whether the provided packet should be kept or dropped."""
        self.packet_count += 1

        if not self._actions:
            return True

        keep, repeat = self._actions[0](self.packet_count)

        if not repeat:
            self._actions.popleft()

        return keep

    def next_packet_duplicated(self) -> bool:
        return False

    def next_packet_out_of_order(self) -> bool:
        return False

    def next_packet_delayed(self) -> bool:
        return False

    def next_packet_dropped(self) -> bool:
        return not self.keep_packet()

    def next_packet_delay(self) -> int:
        return 0

    def next_num_dupes(self) -> int:
        return 0

    def choose_out_of_order_packet(self, max_idx) -> int:
        return 0


class RandomLossGenerator(LossController):
    """Parametrized random number generator that drives a LossyChannel."""

    def __init__(
        self,
        duplicated_packet_probability: float,
        max_duplications_per_packet: int,
        out_of_order_probability: float,
        delayed_packet_probability: float,
        delayed_packet_range_ms: Tuple[int, int],
        dropped_packet_probability: float,
        seed: Optional[int] = None,
    ):
        self.duplicated_packet_probability = duplicated_packet_probability
        self.max_duplications_per_packet = max_duplications_per_packet
        self.out_of_order_probability = out_of_order_probability
        self.delayed_packet_probability = delayed_packet_probability
        self.delayed_packet_range_ms = delayed_packet_range_ms
        self.dropped_packet_probability = dropped_packet_probability
        self._rng = random.Random(seed)

    def next_packet_duplicated(self) -> bool:
        return self.duplicated_packet_probability > self._rng.uniform(0.0, 1.0)

    def next_packet_out_of_order(self) -> bool:
        return self.out_of_order_probability > self._rng.uniform(0.0, 1.0)

    def next_packet_delayed(self) -> bool:
        return self.delayed_packet_probability > self._rng.uniform(0.0, 1.0)

    def next_packet_dropped(self) -> bool:
        return self.dropped_packet_probability > self._rng.uniform(0.0, 1.0)

    def next_packet_delay(self) -> int:
        return self._rng.randint(*self.delayed_packet_range_ms)

    def next_num_dupes(self) -> int:
        return self._rng.randint(1, self.max_duplications_per_packet)

    def choose_out_of_order_packet(self, max_idx) -> int:
        return self._rng.randint(0, max_idx)


class LossyChannel(pw_rpc.ChannelManipulator):
    """Introduces lossy behaviors into a channel."""

    class _Packet:
        """Container class to keep track of incoming packet sequence number."""

        def __init__(self, sequence_number: int, payload: bytes):
            self.sequence_number = sequence_number
            self.payload = payload

    def __init__(
        self, name, loss_generator: LossController, max_num_old_packets=24
    ):
        super().__init__()
        self.name = name
        self._packets: Deque[LossyChannel._Packet] = collections.deque()
        self._old_packets: Deque[LossyChannel._Packet] = collections.deque()
        self._max_old_packet_window_size = max_num_old_packets
        self.unique_packet_count = 0
        self._rng = loss_generator

    def _enqueue_old_packet(self, packet: _Packet):
        if len(self._old_packets) >= self._max_old_packet_window_size:
            self._old_packets.popleft()
        self._old_packets.append(packet)

    def _enqueue_packet(self, payload: bytes):
        # Generate duplicate packets on ingress.
        packet = self._Packet(self.unique_packet_count, payload)
        self.unique_packet_count += 1

        self._packets.append(packet)
        self._enqueue_old_packet(packet)
        if self._rng.next_packet_duplicated():
            num_dupes = self._rng.next_num_dupes()
            _LOG.debug('[%s] Duplicating packet %d times', self.name, num_dupes)
            for _ in range(num_dupes):
                self._packets.append(packet)

    def _send_packets(self):
        while self._packets:
            packet = None

            if self._rng.next_packet_out_of_order():
                idx = self._rng.choose_out_of_order_packet(
                    len(self._old_packets) - 1
                )
                _LOG.debug(
                    '[%s] Selecting out of order packet at index %d',
                    self.name,
                    idx,
                )
                packet = copy.copy(self._old_packets[idx])
                del self._old_packets[idx]
            else:
                packet = self._packets.popleft()

            if self._rng.next_packet_delayed():
                delay = self._rng.next_packet_delay()
                _LOG.debug('[%s] Delaying channel by %d ms', self.name, delay)
                time.sleep(delay / 1000)

            action_msg = 'Dropped'
            if not self._rng.next_packet_dropped():
                action_msg = 'Sent'
                self.send_packet(packet.payload)
            _LOG.debug(
                '[%s] %s packet #%d: %s',
                self.name,
                action_msg,
                packet.sequence_number,
                str(packet.payload),
            )

    def process_and_send(self, packet: bytes):
        self._enqueue_packet(packet)
        self._send_packets()



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Window pane toolbar base class."""

import logging
from typing import Any, Callable, List, Optional
import functools

from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.layout import (
    ConditionalContainer,
    FormattedTextControl,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType

from pw_console.get_pw_console_app import get_pw_console_app
from pw_console.style import (
    get_pane_indicator,
    get_button_style,
    get_toolbar_style,
)
from pw_console.widgets import (
    ToolbarButton,
    mouse_handlers,
    to_checkbox_with_keybind_indicator,
    to_keybind_indicator,
)

_LOG = logging.getLogger(__package__)


class WindowPaneResizeHandle(FormattedTextControl):
    """Button to initiate window pane resize drag events."""

    def __init__(self, parent_window_pane: Any, *args, **kwargs) -> None:
        self.parent_window_pane = parent_window_pane
        super().__init__(*args, **kwargs)

    def mouse_handler(self, mouse_event: MouseEvent):
        """Mouse handler for this control."""
        # Start resize mouse drag event
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            get_pw_console_app().window_manager.start_resize_pane(
                self.parent_window_pane
            )
            # Mouse event handled, return None.
            return None

        # Mouse event not handled, return NotImplemented.
        return NotImplemented


class WindowPaneToolbar:
    """One line toolbar for display at the bottom of of a window pane."""

    # pylint: disable=too-many-instance-attributes
    TOOLBAR_HEIGHT = 1

    def get_left_text_tokens(self):
        """Return toolbar indicator and title."""

        title = self.title
        if not title and self.parent_window_pane:
            # No title was set, fetch the parent window pane title if available.
            parent_pane_title = self.parent_window_pane.pane_title()
            title = parent_pane_title if parent_pane_title else title
        return get_pane_indicator(
            self.focus_check_container, f' {title} ', self.focus_mouse_handler
        )

    def get_center_text_tokens(self):
        """Return formatted text tokens for display in the center part of the
        toolbar."""

        button_style = get_button_style(self.focus_check_container)

        # FormattedTextTuple contents: (Style, Text, Mouse handler)
        separator_text = [('', '  ')]  # 2 spaces of separaton between keybinds.
        if self.focus_mouse_handler:
            separator_text = [('', '  ', self.focus_mouse_handler)]

        fragments = []
        fragments.extend(separator_text)

        for button in self.buttons:
            on_click_handler = None
            if button.mouse_handler:
                on_click_handler = functools.partial(
                    mouse_handlers.on_click,
                    button.mouse_handler,
                )

            if button.is_checkbox:
                fragments.extend(
                    to_checkbox_with_keybind_indicator(
                        button.checked(),
                        button.key,
                        button.description,
                        on_click_handler,
                        base_style=button_style,
                    )
                )
            else:
                fragments.extend(
                    to_keybind_indicator(
                        button.key,
                        button.description,
                        on_click_handler,
                        base_style=button_style,
                    )
                )

            fragments.extend(separator_text)

        # Remaining whitespace should focus on click.
        fragments.extend(separator_text)

        return fragments

    def get_right_text_tokens(self):
        """Return formatted text tokens for display."""
        fragments = []
        if not has_focus(self.focus_check_container.__pt_container__())():
            if self.click_to_focus_text:
                fragments.append(
                    (
                        'class:toolbar-button-inactive '
                        'class:toolbar-button-decoration',
                        ' ',
                        self.focus_mouse_handler,
                    )
                )
                fragments.append(
                    (
                        'class:toolbar-button-inactive class:keyhelp',
                        self.click_to_focus_text,
                        self.focus_mouse_handler,
                    )
                )
                fragments.append(
                    (
                        'class:toolbar-button-inactive '
                        'class:toolbar-button-decoration',
                        ' ',
                        self.focus_mouse_handler,
                    )
                )
        if self.subtitle:
            fragments.append(
                ('', '  {} '.format(self.subtitle()), self.focus_mouse_handler)
            )
        return fragments

    def get_resize_handle(self):
        return get_pane_indicator(
            self.focus_check_container, '─══─', hide_indicator=True
        )

    def add_button(self, button: ToolbarButton):
        self.buttons.append(button)

    def __init__(
        self,
        parent_window_pane: Optional[Any] = None,
        title: Optional[str] = None,
        subtitle: Optional[Callable[[], str]] = None,
        focus_check_container: Optional[Any] = None,
        focus_action_callable: Optional[Callable] = None,
        center_section_align: WindowAlign = WindowAlign.LEFT,
        include_resize_handle: bool = True,
        click_to_focus_text: str = 'click to focus',
    ):
        self.parent_window_pane = parent_window_pane
        self.title = title
        self.subtitle = subtitle
        self.click_to_focus_text = click_to_focus_text

        # Assume check this container for focus
        self.focus_check_container = self
        self.focus_action_callable = None

        # Set parent_window_pane related options
        if self.parent_window_pane:
            if not subtitle:
                self.subtitle = self.parent_window_pane.pane_subtitle
            self.focus_check_container = self.parent_window_pane
            self.focus_action_callable = self.parent_window_pane.focus_self

        # Set title overrides
        if self.subtitle is None:

            def empty_subtitle() -> str:
                return ''

            self.subtitle = empty_subtitle

        if focus_check_container:
            self.focus_check_container = focus_check_container
        if focus_action_callable:
            self.focus_action_callable = focus_action_callable

        self.focus_mouse_handler = None
        if self.focus_action_callable:
            self.focus_mouse_handler = functools.partial(
                mouse_handlers.on_click,
                self.focus_action_callable,
            )

        self.buttons: List[ToolbarButton] = []
        self.show_toolbar = True

        self.left_section_window = Window(
            content=FormattedTextControl(self.get_left_text_tokens),
            align=WindowAlign.LEFT,
            dont_extend_width=True,
        )

        self.center_section_window = Window(
            content=FormattedTextControl(self.get_center_text_tokens),
            align=center_section_align,
            dont_extend_width=False,
        )

        self.right_section_window = Window(
            content=FormattedTextControl(self.get_right_text_tokens),
            # Right side text should appear at the far right of the toolbar
            align=WindowAlign.RIGHT,
            dont_extend_width=True,
        )

        wrapped_get_toolbar_style = functools.partial(
            get_toolbar_style, self.focus_check_container
        )

        sections = [
            self.left_section_window,
            self.center_section_window,
            self.right_section_window,
        ]
        if self.parent_window_pane and include_resize_handle:
            resize_handle = Window(
                content=WindowPaneResizeHandle(
                    self.parent_window_pane,
                    self.get_resize_handle,
                ),
                # Right side text should appear at the far right of the toolbar
                align=WindowAlign.RIGHT,
                dont_extend_width=True,
            )
            sections.append(resize_handle)

        self.toolbar_vsplit = VSplit(
            sections,
            height=WindowPaneToolbar.TOOLBAR_HEIGHT,
            style=wrapped_get_toolbar_style,
        )

        self.container = self._create_toolbar_container(self.toolbar_vsplit)

    def _create_toolbar_container(self, content):
        return ConditionalContainer(
            content, filter=Condition(lambda: self.show_toolbar)
        )

    def __pt_container__(self):
        """Return the prompt_toolkit root container for this log pane.

        This allows self to be used wherever prompt_toolkit expects a container
        object."""
        return self.container  # pylint: disable=no-member



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines a callback-based RPC ClientImpl to use with pw_rpc.Client.

callback_client.Impl supports invoking RPCs synchronously or asynchronously.
Asynchronous invocations use a callback.

Synchronous invocations look like a function call. When invoking a unary or
server streaming RPC, the request may be provided as a message object or as
keyword arguments for the message fields (but not both).

.. code-block:: python

  status, response = client.channel(1).rpcs.MyServer.MyUnary(some_field=123)

  # Calls with a server stream return a status and a list of responses.
  status, responses = rpcs.MyService.MyServerStreaming(Request(some_field=123))

Synchronous client and bidirectional streaming calls accept an iterable of
requests to send.

.. code-block:: python

  requests = [Request(a=1), Request(b=2)]
  status, response = rpcs.MyService.MyClientStreaming(requests)

  requests = [Request(a=1), Request(b=2)]
  status, responses = rpcs.MyService.MyBidirectionalStreaming(requests)

Synchronous invocations block until the RPC completes or times out. The calls
use the default timeout provided when the callback_client.Impl() is created, or
a timeout passed in through the pw_rpc_timeout_s argument. A timeout of None
means to wait indefinitely for a response.

Asynchronous invocations immediately return a call object. Callbacks may be
provided for the three RPC events:

  * on_next(call_object, response) - called for each response
  * on_completed(call_object, status) - called when the RPC completes
  * on_error(call_object, error) - called if the RPC terminates due to an error

The default callbacks simply log the events. If a user-provided callback throws
an exception, that exception is logged and raised when the user calls functions
on the call object.

Unary and client streaming RPCs are invoked asynchronously by calling invoke on
the method object. invoke takes the callbacks. The request may be provided
either as a constructed protobuf or as a dict of proto fields in the
request_args parameter.

.. code-block:: python

  # Pass the request as a protobuf and provide callbacks for all RPC events.
  rpc = client.channel(1).call.MyService.MyServerStreaming
  call = rpc.invoke(rpc.request(a=1), on_next_cb, on_completed_cb, on_error_cb)

  # Create the request from the provided keyword args. Provide a callback for
  # responses, but simply log the other RPC events.
  call = client.channel(1).call.MyServer.MyUnary.invoke(
      request_args=dict(a=1, b=2), on_next=lambda _, reply: process_this(reply))

For client and bidirectional streaming RPCs, requests are sent with the send
method. The finish_and_wait method finishes the client stream. It optionally
takes an iterable for responses to send before closing the stream.

.. code-block:: python

  # Start the call using callbacks.
  call = client.channel(1).rpcs.MyServer.MyClientStream.invoke(on_error=err_cb)

  # Send a single client stream request.
  call.send(some_field=123)

  # Send the requests, close the stream, then wait for the RPC to complete.
  stream_responses = call.finish_and_wait([RequestType(some_field=123), ...])
"""

from pw_rpc.callback_client.call import (
    OptionalTimeout,
    UseDefault,
    UnaryResponse,
    StreamResponse,
    UnaryCall,
    ServerStreamingCall,
    ClientStreamingCall,
    BidirectionalStreamingCall,
    OnNextCallback,
    OnCompletedCallback,
    OnErrorCallback,
)
from pw_rpc.callback_client.errors import RpcError, RpcTimeout
from pw_rpc.callback_client.impl import Impl



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Prints an error message and exits unsuccessfully."""

import argparse
import logging
import os
from pathlib import Path
import subprocess
import sys

try:
    from pw_cli.log import install as setup_logging
except ImportError:
    from logging import basicConfig as setup_logging  # type: ignore

_LOG = logging.getLogger(__name__)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--message', required=True, help='Error message to print'
    )
    parser.add_argument(
        '--target', required=True, help='GN target in which the error occurred'
    )
    parser.add_argument('--root', required=True, type=Path, help='GN root')
    parser.add_argument('--out', required=True, type=Path, help='GN out dir')
    return parser.parse_args()


def main(message: str, target: str, root: Path, out: Path) -> int:
    """Logs the error message and returns 1."""

    _LOG.error('')
    _LOG.error('Build error for %s:', target)
    _LOG.error('')

    for line in message.split('\\n'):
        _LOG.error('  %s', line)

    _LOG.error('')

    gn_cmd = subprocess.run(
        ['gn', 'path', f'--root={root}', out, '//:default', target],
        capture_output=True,
    )
    path_info = gn_cmd.stdout.decode(errors='replace').rstrip()

    relative_out = os.path.relpath(out, root)

    if gn_cmd.returncode == 0 and 'No non-data paths found' not in path_info:
        _LOG.error('Dependency path to this target:')
        _LOG.error('')
        _LOG.error(
            '  gn path %s //:default "%s"\n%s', relative_out, target, path_info
        )
        _LOG.error('')
    else:
        _LOG.error(
            'Run this command to see the build dependency path to this target:'
        )
        _LOG.error('')
        _LOG.error('  gn path %s <target> "%s"', relative_out, target)
        _LOG.error('')
        _LOG.error('where <target> is the GN target you are building.')
        _LOG.error('')

    return 1


if __name__ == '__main__':
    setup_logging()
    sys.exit(main(**vars(_parse_args())))



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of the Raspberry Pi Pico SDK."""

from contextlib import contextmanager
import os
from pathlib import Path
from typing import Sequence
import subprocess

import pw_package.git_repo
import pw_package.package_manager


@contextmanager
def change_working_dir(directory: Path):
    original_dir = Path.cwd()
    try:
        os.chdir(directory)
        yield directory
    finally:
        os.chdir(original_dir)


class PiPicoSdk(pw_package.package_manager.Package):
    """Install and check status of the Raspberry Pi Pico SDK."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='pico_sdk', **kwargs)
        self._pico_sdk = pw_package.git_repo.GitRepo(
            name='pico_sdk',
            url='https://github.com/raspberrypi/pico-sdk',
            commit='2e6142b15b8a75c1227dd3edbe839193b2bf9041',
        )

    def install(self, path: Path) -> None:
        self._pico_sdk.install(path)

        # Run submodule update --init to fetch tinyusb.
        with change_working_dir(path) as _pico_sdk_repo:
            subprocess.run(
                ['git', 'submodule', 'update', '--init'], capture_output=True
            )

    def info(self, path: Path) -> Sequence[str]:
        return (
            f'{self.name} installed in: {path}',
            "Enable by running 'gn args out' and adding this line:",
            f'  PICO_SRC_DIR = "{path}"',
        )


pw_package.package_manager.register(PiPicoSdk)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools to retrieve and parse metrics."""
from collections import defaultdict
import json
import logging
from typing import Any, Optional
from pw_tokenizer import detokenize

_LOG = logging.getLogger(__name__)


def _tree():
    """Creates a key based on given input."""
    return defaultdict(_tree)


def _insert(metrics, path_names, value):
    """Inserts any value in a leaf of the dictionary."""
    for index, path_name in enumerate(path_names):
        if index < len(path_names) - 1:
            metrics = metrics[path_name]
        elif path_name in metrics:
            # the value in this position isn't a float or int,
            # then collision occurs, throw an error.
            assert ValueError(
                'Variable already exists: {p}'.format(p=path_name)
            )
        else:
            metrics[path_name] = value


def parse_metrics(
    rpcs: Any,
    detokenizer: Optional[detokenize.Detokenizer],
    timeout_s: Optional[float],
):
    """Detokenizes metric names and retrieves their values."""
    # Creates a defaultdict that can infinitely have other defaultdicts
    # without a specified type.
    metrics: defaultdict = _tree()
    if not detokenizer:
        _LOG.error('No metrics token database set.')
        return metrics
    stream_response = rpcs.pw.metric.proto.MetricService.Get(
        pw_rpc_timeout_s=timeout_s
    )
    if not stream_response.status.ok():
        _LOG.error('Unexpected status %s', stream_response.status)
        return metrics
    for metric_response in stream_response.responses:
        for metric in metric_response.metrics:
            path_names = []
            for path in metric.token_path:
                path_name = str(
                    detokenize.DetokenizedString(
                        path, detokenizer.lookup(path), b'', False
                    )
                ).strip('"')
                path_names.append(path_name)
            value = (
                metric.as_float
                if metric.HasField('as_float')
                else metric.as_int
            )
            # inserting path_names into metrics.
            _insert(metrics, path_names, value)
    # Converts default dict objects into standard dictionaries.
    return json.loads(json.dumps(metrics))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_metric_proto/metric_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_metric_proto/metric_service.proto',
  package='pw.metric.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n$pw_metric_proto/metric_service.proto\x12\x0fpw.metric.proto\"`\n\x06Metric\x12\x12\n\ntoken_path\x18\x01 \x03(\x07\x12\x13\n\x0bstring_path\x18\x02 \x03(\t\x12\x12\n\x08\x61s_float\x18\x03 \x01(\x02H\x00\x12\x10\n\x06\x61s_int\x18\x04 \x01(\rH\x00\x42\x07\n\x05value\"9\n\rMetricRequest\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric\":\n\x0eMetricResponse\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric2[\n\rMetricService\x12J\n\x03Get\x12\x1e.pw.metric.proto.MetricRequest\x1a\x1f.pw.metric.proto.MetricResponse\"\x00\x30\x01\x62\x06proto3'
)




_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='pw.metric.proto.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_path', full_name='pw.metric.proto.Metric.token_path', index=0,
      number=1, type=7, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_path', full_name='pw.metric.proto.Metric.string_path', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='as_float', full_name='pw.metric.proto.Metric.as_float', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='as_int', full_name='pw.metric.proto.Metric.as_int', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='pw.metric.proto.Metric.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=57,
  serialized_end=153,
)


_METRICREQUEST = _descriptor.Descriptor(
  name='MetricRequest',
  full_name='pw.metric.proto.MetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='pw.metric.proto.MetricRequest.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=212,
)


_METRICRESPONSE = _descriptor.Descriptor(
  name='MetricResponse',
  full_name='pw.metric.proto.MetricResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='pw.metric.proto.MetricResponse.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=272,
)

_METRIC.oneofs_by_name['value'].fields.append(
  _METRIC.fields_by_name['as_float'])
_METRIC.fields_by_name['as_float'].containing_oneof = _METRIC.oneofs_by_name['value']
_METRIC.oneofs_by_name['value'].fields.append(
  _METRIC.fields_by_name['as_int'])
_METRIC.fields_by_name['as_int'].containing_oneof = _METRIC.oneofs_by_name['value']
_METRICREQUEST.fields_by_name['metrics'].message_type = _METRIC
_METRICRESPONSE.fields_by_name['metrics'].message_type = _METRIC
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['MetricRequest'] = _METRICREQUEST
DESCRIPTOR.message_types_by_name['MetricResponse'] = _METRICRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.Metric)
  })
_sym_db.RegisterMessage(Metric)

MetricRequest = _reflection.GeneratedProtocolMessageType('MetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _METRICREQUEST,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.MetricRequest)
  })
_sym_db.RegisterMessage(MetricRequest)

MetricResponse = _reflection.GeneratedProtocolMessageType('MetricResponse', (_message.Message,), {
  'DESCRIPTOR' : _METRICRESPONSE,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.MetricResponse)
  })
_sym_db.RegisterMessage(MetricResponse)



_METRICSERVICE = _descriptor.ServiceDescriptor(
  name='MetricService',
  full_name='pw.metric.proto.MetricService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=274,
  serialized_end=365,
  methods=[
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='pw.metric.proto.MetricService.Get',
    index=0,
    containing_service=None,
    input_type=_METRICREQUEST,
    output_type=_METRICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_METRICSERVICE)

DESCRIPTOR.services_by_name['MetricService'] = _METRICSERVICE

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/tuf.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/tuf.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cpw_software_update/tuf.proto\x12\x12pw.software_update\x1a\x1fgoogle/protobuf/timestamp.proto\"i\n\x12SignedRootMetadata\x12 \n\x18serialized_root_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"s\n\x17SignedTimestampMetadata\x12%\n\x1dserialized_timestamp_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"q\n\x16SignedSnapshotMetadata\x12$\n\x1cserialized_snapshot_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"o\n\x15SignedTargetsMetadata\x12#\n\x1bserialized_targets_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"\x91\x01\n\x0e\x43ommonMetadata\x12\x14\n\x0cspec_version\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x30\n\x07\x65xpires\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x12\x11\n\x04role\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_expiresB\x07\n\x05_role\"\xed\x03\n\x0cRootMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x1b\n\x13\x63onsistent_snapshot\x18\x02 \x01(\x08\x12,\n\x04keys\x18\x03 \x03(\x0b\x32\x1e.pw.software_update.KeyMapping\x12L\n\x1aroot_signature_requirement\x18\x04 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12Q\n\x1ftimestamp_signature_requirement\x18\x05 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12P\n\x1esnapshot_signature_requirement\x18\x06 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12O\n\x1dtargets_signature_requirement\x18\x07 \x01(\x0b\x32(.pw.software_update.SignatureRequirementJ\x04\x08\x08\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\xa0\x01\n\x11TimestampMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12;\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x9e\x01\n\x10SnapshotMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12:\n\x10targets_metadata\x18\x02 \x03(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x97\x01\n\x0fTargetsMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x34\n\x0ctarget_files\x18\x02 \x03(\x0b\x32\x1e.pw.software_update.TargetFileJ\x04\x08\t\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"(\n\tSignature\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12\x0b\n\x03sig\x18\x02 \x01(\x0c\"B\n\nKeyMapping\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12$\n\x03key\x18\x02 \x01(\x0b\x32\x17.pw.software_update.Key\"s\n\x03Key\x12-\n\x08key_type\x18\x01 \x01(\x0e\x32\x1b.pw.software_update.KeyType\x12-\n\x06scheme\x18\x02 \x01(\x0e\x32\x1d.pw.software_update.KeyScheme\x12\x0e\n\x06keyval\x18\x03 \x01(\x0c\":\n\x14SignatureRequirement\x12\x0f\n\x07key_ids\x18\x01 \x03(\x0c\x12\x11\n\tthreshold\x18\x02 \x01(\r\"H\n\x04Hash\x12\x32\n\x08\x66unction\x18\x01 \x01(\x0e\x32 .pw.software_update.HashFunction\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"l\n\nTargetFile\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x0e\n\x06length\x18\x02 \x01(\x04\x12(\n\x06hashes\x18\x03 \x03(\x0b\x32\x18.pw.software_update.HashJ\x04\x08\x04\x10\x10J\x04\x08\x10\x10 J\x05\x08 \x10\x80\x02\"\x8f\x01\n\x0cMetadataFile\x12\x16\n\tfile_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x13\n\x06length\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12(\n\x06hashes\x18\x04 \x03(\x0b\x32\x18.pw.software_update.HashB\x0c\n\n_file_nameB\t\n\x07_length*5\n\x0cHashFunction\x12\x19\n\x15UNKNOWN_HASH_FUNCTION\x10\x00\x12\n\n\x06SHA256\x10\x01*N\n\x07KeyType\x12\x14\n\x10UNKNOWN_KEY_TYPE\x10\x00\x12\x07\n\x03RSA\x10\x01\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x02\x12\x17\n\x13\x45\x43\x44SA_SHA2_NISTP256\x10\x03*u\n\tKeyScheme\x12\x16\n\x12UNKNOWN_KEY_SCHEME\x10\x00\x12\x1c\n\x18RSASSA_PSS_SHA256_SCHEME\x10\x01\x12\x12\n\x0e\x45\x44\x32\x35\x35\x31\x39_SCHEME\x10\x02\x12\x1e\n\x1a\x45\x43\x44SA_SHA2_NISTP256_SCHEME\x10\x03\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_HASHFUNCTION = _descriptor.EnumDescriptor(
  name='HashFunction',
  full_name='pw.software_update.HashFunction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_HASH_FUNCTION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHA256', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2276,
  serialized_end=2329,
)
_sym_db.RegisterEnumDescriptor(_HASHFUNCTION)

HashFunction = enum_type_wrapper.EnumTypeWrapper(_HASHFUNCTION)
_KEYTYPE = _descriptor.EnumDescriptor(
  name='KeyType',
  full_name='pw.software_update.KeyType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KEY_TYPE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RSA', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ED25519', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ECDSA_SHA2_NISTP256', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2331,
  serialized_end=2409,
)
_sym_db.RegisterEnumDescriptor(_KEYTYPE)

KeyType = enum_type_wrapper.EnumTypeWrapper(_KEYTYPE)
_KEYSCHEME = _descriptor.EnumDescriptor(
  name='KeyScheme',
  full_name='pw.software_update.KeyScheme',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KEY_SCHEME', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RSASSA_PSS_SHA256_SCHEME', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ED25519_SCHEME', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ECDSA_SHA2_NISTP256_SCHEME', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2411,
  serialized_end=2528,
)
_sym_db.RegisterEnumDescriptor(_KEYSCHEME)

KeyScheme = enum_type_wrapper.EnumTypeWrapper(_KEYSCHEME)
UNKNOWN_HASH_FUNCTION = 0
SHA256 = 1
UNKNOWN_KEY_TYPE = 0
RSA = 1
ED25519 = 2
ECDSA_SHA2_NISTP256 = 3
UNKNOWN_KEY_SCHEME = 0
RSASSA_PSS_SHA256_SCHEME = 1
ED25519_SCHEME = 2
ECDSA_SHA2_NISTP256_SCHEME = 3



_SIGNEDROOTMETADATA = _descriptor.Descriptor(
  name='SignedRootMetadata',
  full_name='pw.software_update.SignedRootMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_root_metadata', full_name='pw.software_update.SignedRootMetadata.serialized_root_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedRootMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=190,
)


_SIGNEDTIMESTAMPMETADATA = _descriptor.Descriptor(
  name='SignedTimestampMetadata',
  full_name='pw.software_update.SignedTimestampMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_timestamp_metadata', full_name='pw.software_update.SignedTimestampMetadata.serialized_timestamp_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedTimestampMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=192,
  serialized_end=307,
)


_SIGNEDSNAPSHOTMETADATA = _descriptor.Descriptor(
  name='SignedSnapshotMetadata',
  full_name='pw.software_update.SignedSnapshotMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_snapshot_metadata', full_name='pw.software_update.SignedSnapshotMetadata.serialized_snapshot_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedSnapshotMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=422,
)


_SIGNEDTARGETSMETADATA = _descriptor.Descriptor(
  name='SignedTargetsMetadata',
  full_name='pw.software_update.SignedTargetsMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_targets_metadata', full_name='pw.software_update.SignedTargetsMetadata.serialized_targets_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedTargetsMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=424,
  serialized_end=535,
)


_COMMONMETADATA = _descriptor.Descriptor(
  name='CommonMetadata',
  full_name='pw.software_update.CommonMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spec_version', full_name='pw.software_update.CommonMetadata.spec_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='pw.software_update.CommonMetadata.version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expires', full_name='pw.software_update.CommonMetadata.expires', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='pw.software_update.CommonMetadata.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_expires', full_name='pw.software_update.CommonMetadata._expires',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_role', full_name='pw.software_update.CommonMetadata._role',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=538,
  serialized_end=683,
)


_ROOTMETADATA = _descriptor.Descriptor(
  name='RootMetadata',
  full_name='pw.software_update.RootMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.RootMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='consistent_snapshot', full_name='pw.software_update.RootMetadata.consistent_snapshot', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keys', full_name='pw.software_update.RootMetadata.keys', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='root_signature_requirement', full_name='pw.software_update.RootMetadata.root_signature_requirement', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_signature_requirement', full_name='pw.software_update.RootMetadata.timestamp_signature_requirement', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_signature_requirement', full_name='pw.software_update.RootMetadata.snapshot_signature_requirement', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_signature_requirement', full_name='pw.software_update.RootMetadata.targets_signature_requirement', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=686,
  serialized_end=1179,
)


_TIMESTAMPMETADATA = _descriptor.Descriptor(
  name='TimestampMetadata',
  full_name='pw.software_update.TimestampMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.TimestampMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_metadata', full_name='pw.software_update.TimestampMetadata.snapshot_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1182,
  serialized_end=1342,
)


_SNAPSHOTMETADATA = _descriptor.Descriptor(
  name='SnapshotMetadata',
  full_name='pw.software_update.SnapshotMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.SnapshotMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.SnapshotMetadata.targets_metadata', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1345,
  serialized_end=1503,
)


_TARGETSMETADATA = _descriptor.Descriptor(
  name='TargetsMetadata',
  full_name='pw.software_update.TargetsMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.TargetsMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_files', full_name='pw.software_update.TargetsMetadata.target_files', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1506,
  serialized_end=1657,
)


_SIGNATURE = _descriptor.Descriptor(
  name='Signature',
  full_name='pw.software_update.Signature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_id', full_name='pw.software_update.Signature.key_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sig', full_name='pw.software_update.Signature.sig', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1659,
  serialized_end=1699,
)


_KEYMAPPING = _descriptor.Descriptor(
  name='KeyMapping',
  full_name='pw.software_update.KeyMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_id', full_name='pw.software_update.KeyMapping.key_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.KeyMapping.key', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1701,
  serialized_end=1767,
)


_KEY = _descriptor.Descriptor(
  name='Key',
  full_name='pw.software_update.Key',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_type', full_name='pw.software_update.Key.key_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheme', full_name='pw.software_update.Key.scheme', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keyval', full_name='pw.software_update.Key.keyval', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1769,
  serialized_end=1884,
)


_SIGNATUREREQUIREMENT = _descriptor.Descriptor(
  name='SignatureRequirement',
  full_name='pw.software_update.SignatureRequirement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_ids', full_name='pw.software_update.SignatureRequirement.key_ids', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threshold', full_name='pw.software_update.SignatureRequirement.threshold', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1886,
  serialized_end=1944,
)


_HASH = _descriptor.Descriptor(
  name='Hash',
  full_name='pw.software_update.Hash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='pw.software_update.Hash.function', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pw.software_update.Hash.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1946,
  serialized_end=2018,
)


_TARGETFILE = _descriptor.Descriptor(
  name='TargetFile',
  full_name='pw.software_update.TargetFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.software_update.TargetFile.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='pw.software_update.TargetFile.length', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashes', full_name='pw.software_update.TargetFile.hashes', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2020,
  serialized_end=2128,
)


_METADATAFILE = _descriptor.Descriptor(
  name='MetadataFile',
  full_name='pw.software_update.MetadataFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.software_update.MetadataFile.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='pw.software_update.MetadataFile.version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='pw.software_update.MetadataFile.length', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashes', full_name='pw.software_update.MetadataFile.hashes', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_file_name', full_name='pw.software_update.MetadataFile._file_name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_length', full_name='pw.software_update.MetadataFile._length',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2131,
  serialized_end=2274,
)

_SIGNEDROOTMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDTIMESTAMPMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDSNAPSHOTMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDTARGETSMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_COMMONMETADATA.fields_by_name['expires'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COMMONMETADATA.oneofs_by_name['_expires'].fields.append(
  _COMMONMETADATA.fields_by_name['expires'])
_COMMONMETADATA.fields_by_name['expires'].containing_oneof = _COMMONMETADATA.oneofs_by_name['_expires']
_COMMONMETADATA.oneofs_by_name['_role'].fields.append(
  _COMMONMETADATA.fields_by_name['role'])
_COMMONMETADATA.fields_by_name['role'].containing_oneof = _COMMONMETADATA.oneofs_by_name['_role']
_ROOTMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_ROOTMETADATA.fields_by_name['keys'].message_type = _KEYMAPPING
_ROOTMETADATA.fields_by_name['root_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['timestamp_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['snapshot_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['targets_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_TIMESTAMPMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_TIMESTAMPMETADATA.fields_by_name['snapshot_metadata'].message_type = _METADATAFILE
_SNAPSHOTMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_SNAPSHOTMETADATA.fields_by_name['targets_metadata'].message_type = _METADATAFILE
_TARGETSMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_TARGETSMETADATA.fields_by_name['target_files'].message_type = _TARGETFILE
_KEYMAPPING.fields_by_name['key'].message_type = _KEY
_KEY.fields_by_name['key_type'].enum_type = _KEYTYPE
_KEY.fields_by_name['scheme'].enum_type = _KEYSCHEME
_HASH.fields_by_name['function'].enum_type = _HASHFUNCTION
_TARGETFILE.fields_by_name['hashes'].message_type = _HASH
_METADATAFILE.fields_by_name['hashes'].message_type = _HASH
_METADATAFILE.oneofs_by_name['_file_name'].fields.append(
  _METADATAFILE.fields_by_name['file_name'])
_METADATAFILE.fields_by_name['file_name'].containing_oneof = _METADATAFILE.oneofs_by_name['_file_name']
_METADATAFILE.oneofs_by_name['_length'].fields.append(
  _METADATAFILE.fields_by_name['length'])
_METADATAFILE.fields_by_name['length'].containing_oneof = _METADATAFILE.oneofs_by_name['_length']
DESCRIPTOR.message_types_by_name['SignedRootMetadata'] = _SIGNEDROOTMETADATA
DESCRIPTOR.message_types_by_name['SignedTimestampMetadata'] = _SIGNEDTIMESTAMPMETADATA
DESCRIPTOR.message_types_by_name['SignedSnapshotMetadata'] = _SIGNEDSNAPSHOTMETADATA
DESCRIPTOR.message_types_by_name['SignedTargetsMetadata'] = _SIGNEDTARGETSMETADATA
DESCRIPTOR.message_types_by_name['CommonMetadata'] = _COMMONMETADATA
DESCRIPTOR.message_types_by_name['RootMetadata'] = _ROOTMETADATA
DESCRIPTOR.message_types_by_name['TimestampMetadata'] = _TIMESTAMPMETADATA
DESCRIPTOR.message_types_by_name['SnapshotMetadata'] = _SNAPSHOTMETADATA
DESCRIPTOR.message_types_by_name['TargetsMetadata'] = _TARGETSMETADATA
DESCRIPTOR.message_types_by_name['Signature'] = _SIGNATURE
DESCRIPTOR.message_types_by_name['KeyMapping'] = _KEYMAPPING
DESCRIPTOR.message_types_by_name['Key'] = _KEY
DESCRIPTOR.message_types_by_name['SignatureRequirement'] = _SIGNATUREREQUIREMENT
DESCRIPTOR.message_types_by_name['Hash'] = _HASH
DESCRIPTOR.message_types_by_name['TargetFile'] = _TARGETFILE
DESCRIPTOR.message_types_by_name['MetadataFile'] = _METADATAFILE
DESCRIPTOR.enum_types_by_name['HashFunction'] = _HASHFUNCTION
DESCRIPTOR.enum_types_by_name['KeyType'] = _KEYTYPE
DESCRIPTOR.enum_types_by_name['KeyScheme'] = _KEYSCHEME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SignedRootMetadata = _reflection.GeneratedProtocolMessageType('SignedRootMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDROOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedRootMetadata)
  })
_sym_db.RegisterMessage(SignedRootMetadata)

SignedTimestampMetadata = _reflection.GeneratedProtocolMessageType('SignedTimestampMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDTIMESTAMPMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedTimestampMetadata)
  })
_sym_db.RegisterMessage(SignedTimestampMetadata)

SignedSnapshotMetadata = _reflection.GeneratedProtocolMessageType('SignedSnapshotMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDSNAPSHOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedSnapshotMetadata)
  })
_sym_db.RegisterMessage(SignedSnapshotMetadata)

SignedTargetsMetadata = _reflection.GeneratedProtocolMessageType('SignedTargetsMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDTARGETSMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedTargetsMetadata)
  })
_sym_db.RegisterMessage(SignedTargetsMetadata)

CommonMetadata = _reflection.GeneratedProtocolMessageType('CommonMetadata', (_message.Message,), {
  'DESCRIPTOR' : _COMMONMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.CommonMetadata)
  })
_sym_db.RegisterMessage(CommonMetadata)

RootMetadata = _reflection.GeneratedProtocolMessageType('RootMetadata', (_message.Message,), {
  'DESCRIPTOR' : _ROOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.RootMetadata)
  })
_sym_db.RegisterMessage(RootMetadata)

TimestampMetadata = _reflection.GeneratedProtocolMessageType('TimestampMetadata', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TimestampMetadata)
  })
_sym_db.RegisterMessage(TimestampMetadata)

SnapshotMetadata = _reflection.GeneratedProtocolMessageType('SnapshotMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SnapshotMetadata)
  })
_sym_db.RegisterMessage(SnapshotMetadata)

TargetsMetadata = _reflection.GeneratedProtocolMessageType('TargetsMetadata', (_message.Message,), {
  'DESCRIPTOR' : _TARGETSMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TargetsMetadata)
  })
_sym_db.RegisterMessage(TargetsMetadata)

Signature = _reflection.GeneratedProtocolMessageType('Signature', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATURE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Signature)
  })
_sym_db.RegisterMessage(Signature)

KeyMapping = _reflection.GeneratedProtocolMessageType('KeyMapping', (_message.Message,), {
  'DESCRIPTOR' : _KEYMAPPING,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.KeyMapping)
  })
_sym_db.RegisterMessage(KeyMapping)

Key = _reflection.GeneratedProtocolMessageType('Key', (_message.Message,), {
  'DESCRIPTOR' : _KEY,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Key)
  })
_sym_db.RegisterMessage(Key)

SignatureRequirement = _reflection.GeneratedProtocolMessageType('SignatureRequirement', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATUREREQUIREMENT,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignatureRequirement)
  })
_sym_db.RegisterMessage(SignatureRequirement)

Hash = _reflection.GeneratedProtocolMessageType('Hash', (_message.Message,), {
  'DESCRIPTOR' : _HASH,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Hash)
  })
_sym_db.RegisterMessage(Hash)

TargetFile = _reflection.GeneratedProtocolMessageType('TargetFile', (_message.Message,), {
  'DESCRIPTOR' : _TARGETFILE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TargetFile)
  })
_sym_db.RegisterMessage(TargetFile)

MetadataFile = _reflection.GeneratedProtocolMessageType('MetadataFile', (_message.Message,), {
  'DESCRIPTOR' : _METADATAFILE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.MetadataFile)
  })
_sym_db.RegisterMessage(MetadataFile)


# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Script for generating test bundles"""

import argparse
import subprocess
import sys
from typing import Dict, Optional

from pw_software_update import dev_sign, keys, metadata, root_metadata
from pw_software_update.update_bundle_pb2 import Manifest, UpdateBundle
from pw_software_update.tuf_pb2 import (
    RootMetadata,
    SignedRootMetadata,
    TargetsMetadata,
    SignedTargetsMetadata,
)

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization

HEADER = """// Copyright 2021 The Pigweed Authors
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations under
// the License.

#pragma once

#include "pw_bytes/span.h"

"""

TEST_DEV_KEY = """-----BEGIN PRIVATE KEY-----
MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgVgMQBOTJyx1xOafy
WTs2VkACf7Uo3RbP9Vun+oKXtMihRANCAATV7XJljxeUs2z2wqM5Q/kohAra1620
zXT90N9a3UR+IHksTd1OA7wFq220IQB/e4eVzbcOprN0MMMuSgXMxL8p
-----END PRIVATE KEY-----"""

TEST_PROD_KEY = """-----BEGIN PRIVATE KEY-----
MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQg73MLNmB/fPNX75Pl
YdynPtJkM2gGOWfIcHDuwuxSQmqhRANCAARpvjrXkjG2Fp+ZgREtxeTBBmJmWGS9
8Ny2tXY+Qggzl77G7wvCNF5+koz7ecsV6sKjK+dFiAXOIdqlga7p2j0A
-----END PRIVATE KEY-----"""

TEST_TARGETS_DEV_KEY = """-----BEGIN PRIVATE KEY-----
MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQggRCrido5vZOnkULH
sxQDt9Qoe/TlEKoqa1bhO1HFbi6hRANCAASVwdXbGWM7+f/r+Z2W6Dbd7CQA0Cbb
pkBv5PnA+DZnCkFhLW2kTn89zQv8W1x4m9maoINp9QPXQ4/nXlrVHqDg
-----END PRIVATE KEY-----"""

TEST_TARGETS_PROD_KEY = """-----BEGIN PRIVATE KEY-----
MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgx2VdB2EsUKghuLMG
RmxzqX2jnLTq5pxsFgO5Rrf5jlehRANCAASVijeDpemxVSlgZOOW0yvwE5QkXkq0
geWonkusMP0+MXopnmN0QlpgaCnG40TSr/W+wFjRmNCklL4dXk01oCwD
-----END PRIVATE KEY-----"""

TEST_ROOT_VERSION = 2
TEST_TARGETS_VERSION = 2

USER_MANIFEST_FILE_NAME = 'user_manifest'

TARGET_FILES = {
    'file1': 'file 1 content'.encode(),
    'file2': 'file 2 content'.encode(),
    USER_MANIFEST_FILE_NAME: 'user manfiest content'.encode(),
}


def byte_array_declaration(data: bytes, name: str) -> str:
    """Generates a byte C array declaration for a byte array"""
    type_name = '[[maybe_unused]] const std::byte'
    byte_str = ''.join([f'std::byte{{0x{b:02x}}},' for b in data])
    array_body = f'{{{byte_str}}}'
    return f'{type_name} {name}[] = {array_body};'


def proto_array_declaration(proto, name: str) -> str:
    """Generates a byte array declaration for a proto"""
    return byte_array_declaration(proto.SerializeToString(), name)


def private_key_public_pem_bytes(key: ec.EllipticCurvePrivateKey) -> bytes:
    """Serializes the public part of a private key in PEM format"""
    return key.public_key().public_bytes(
        serialization.Encoding.PEM,
        serialization.PublicFormat.SubjectPublicKeyInfo,
    )


def private_key_private_pem_bytes(key: ec.EllipticCurvePrivateKey) -> bytes:
    """Serializes the private part of a private key in PEM format"""
    return key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


class Bundle:
    """A helper for test UpdateBundle generation"""

    def __init__(self):
        self._root_dev_key = serialization.load_pem_private_key(
            TEST_DEV_KEY.encode(), None
        )
        self._root_prod_key = serialization.load_pem_private_key(
            TEST_PROD_KEY.encode(), None
        )
        self._targets_dev_key = serialization.load_pem_private_key(
            TEST_TARGETS_DEV_KEY.encode(), None
        )
        self._targets_prod_key = serialization.load_pem_private_key(
            TEST_TARGETS_PROD_KEY.encode(), None
        )
        self._payloads: Dict[str, bytes] = {}
        # Adds some update files.
        for key, value in TARGET_FILES.items():
            self.add_payload(key, value)

    def add_payload(self, name: str, payload: bytes) -> None:
        """Adds a payload to the bundle"""
        self._payloads[name] = payload

    def generate_dev_root_metadata(self) -> RootMetadata:
        """Generates a root metadata with the dev key"""
        # The dev root metadata contains both the prod and the dev public key,
        # so that it can rotate to prod. But it will only use a dev targets
        # key.
        return root_metadata.gen_root_metadata(
            root_metadata.RootKeys(
                [
                    private_key_public_pem_bytes(self._root_dev_key),
                    private_key_public_pem_bytes(self._root_prod_key),
                ]
            ),
            root_metadata.TargetsKeys(
                [private_key_public_pem_bytes(self._targets_dev_key)]
            ),
            TEST_ROOT_VERSION,
        )

    def generate_prod_root_metadata(self) -> RootMetadata:
        """Generates a root metadata with the prod key"""
        # The prod root metadta contains only the prod public key and uses the
        # prod targets key
        return root_metadata.gen_root_metadata(
            root_metadata.RootKeys(
                [private_key_public_pem_bytes(self._root_prod_key)]
            ),
            root_metadata.TargetsKeys(
                [private_key_public_pem_bytes(self._targets_prod_key)]
            ),
            TEST_ROOT_VERSION,
        )

    def generate_dev_signed_root_metadata(self) -> SignedRootMetadata:
        """Generates a dev signed root metadata"""
        signed_root = SignedRootMetadata()
        root_metadata_proto = self.generate_dev_root_metadata()
        signed_root.serialized_root_metadata = (
            root_metadata_proto.SerializeToString()
        )
        return dev_sign.sign_root_metadata(
            signed_root, private_key_private_pem_bytes(self._root_dev_key)
        )

    def generate_prod_signed_root_metadata(
        self, root_metadata_proto: Optional[RootMetadata] = None
    ) -> SignedRootMetadata:
        """Generates a root metadata signed by the prod key"""
        if not root_metadata_proto:
            root_metadata_proto = self.generate_prod_root_metadata()

        signed_root = SignedRootMetadata(
            serialized_root_metadata=root_metadata_proto.SerializeToString()
        )

        return dev_sign.sign_root_metadata(
            signed_root, private_key_private_pem_bytes(self._root_prod_key)
        )

    def generate_targets_metadata(self) -> TargetsMetadata:
        """Generates the targets metadata"""
        targets = metadata.gen_targets_metadata(
            self._payloads, metadata.DEFAULT_HASHES, TEST_TARGETS_VERSION
        )
        return targets

    def generate_unsigned_bundle(
        self,
        targets_metadata: Optional[TargetsMetadata] = None,
        signed_root_metadata: Optional[SignedRootMetadata] = None,
    ) -> UpdateBundle:
        """Generate an unsigned (targets metadata) update bundle"""
        bundle = UpdateBundle()

        if not targets_metadata:
            targets_metadata = self.generate_targets_metadata()

        if signed_root_metadata:
            bundle.root_metadata.CopyFrom(signed_root_metadata)

        bundle.targets_metadata['targets'].CopyFrom(
            SignedTargetsMetadata(
                serialized_targets_metadata=targets_metadata.SerializeToString()
            )
        )

        for name, payload in self._payloads.items():
            bundle.target_payloads[name] = payload

        return bundle

    def generate_dev_signed_bundle(
        self,
        targets_metadata_override: Optional[TargetsMetadata] = None,
        signed_root_metadata: Optional[SignedRootMetadata] = None,
    ) -> UpdateBundle:
        """Generate a dev signed update bundle"""
        return dev_sign.sign_update_bundle(
            self.generate_unsigned_bundle(
                targets_metadata_override, signed_root_metadata
            ),
            private_key_private_pem_bytes(self._targets_dev_key),
        )

    def generate_prod_signed_bundle(
        self,
        targets_metadata_override: Optional[TargetsMetadata] = None,
        signed_root_metadata: Optional[SignedRootMetadata] = None,
    ) -> UpdateBundle:
        """Generate a prod signed update bundle"""
        # The targets metadata in a prod signed bundle can only be verified
        # by a prod signed root. Because it is signed by the prod targets key.
        # The prod signed root however, can be verified by a dev root.
        return dev_sign.sign_update_bundle(
            self.generate_unsigned_bundle(
                targets_metadata_override, signed_root_metadata
            ),
            private_key_private_pem_bytes(self._targets_prod_key),
        )

    def generate_manifest(self) -> Manifest:
        """Generates the manifest"""
        manifest = Manifest()
        manifest.targets_metadata['targets'].CopyFrom(
            self.generate_targets_metadata()
        )
        if USER_MANIFEST_FILE_NAME in self._payloads:
            manifest.user_manifest = self._payloads[USER_MANIFEST_FILE_NAME]
        return manifest


def parse_args():
    """Setup argparse."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "output_header", help="output path of the generated C header"
    )
    return parser.parse_args()


# TODO(b/237580538): Refactor the code so that each test bundle generation
# is done in a separate function or script.
# pylint: disable=too-many-locals
def main() -> int:
    """Main"""
    args = parse_args()

    test_bundle = Bundle()

    dev_signed_root = test_bundle.generate_dev_signed_root_metadata()
    dev_signed_bundle = test_bundle.generate_dev_signed_bundle()
    dev_signed_bundle_with_root = test_bundle.generate_dev_signed_bundle(
        signed_root_metadata=dev_signed_root
    )
    unsigned_bundle_with_root = test_bundle.generate_unsigned_bundle(
        signed_root_metadata=dev_signed_root
    )
    manifest_proto = test_bundle.generate_manifest()
    prod_signed_root = test_bundle.generate_prod_signed_root_metadata()
    prod_signed_bundle = test_bundle.generate_prod_signed_bundle(
        None, prod_signed_root
    )
    dev_signed_bundle_with_prod_root = test_bundle.generate_dev_signed_bundle(
        signed_root_metadata=prod_signed_root
    )

    # Generates a prod root metadata that fails signature verification against
    # the dev root (i.e. it has a bad prod signature). This is done by making
    # a bad prod signature.
    bad_prod_signature = test_bundle.generate_prod_root_metadata()
    signed_bad_prod_signature = test_bundle.generate_prod_signed_root_metadata(
        bad_prod_signature
    )
    # Compromises the signature.
    signed_bad_prod_signature.signatures[0].sig = b'1' * 64
    signed_bad_prod_signature_bundle = test_bundle.generate_prod_signed_bundle(
        None, signed_bad_prod_signature
    )

    # Generates a prod root metadtata that fails to verify itself. Specifically,
    # the prod signature cannot be verified by the key in the incoming root
    # metadata. This is done by dev signing a prod root metadata.
    # pylint: disable=line-too-long
    signed_mismatched_root_key_and_signature = SignedRootMetadata(
        serialized_root_metadata=test_bundle.generate_prod_root_metadata().SerializeToString()
    )
    # pylint: enable=line-too-long
    dev_root_key = serialization.load_pem_private_key(
        TEST_DEV_KEY.encode(), None
    )
    signature = keys.create_ecdsa_signature(
        signed_mismatched_root_key_and_signature.serialized_root_metadata,
        private_key_private_pem_bytes(dev_root_key),  # type: ignore
    )
    signed_mismatched_root_key_and_signature.signatures.append(signature)
    mismatched_root_key_and_signature_bundle = (
        test_bundle.generate_prod_signed_bundle(
            None, signed_mismatched_root_key_and_signature
        )
    )

    # Generates a prod root metadata with rollback attempt.
    root_rollback = test_bundle.generate_prod_root_metadata()
    root_rollback.common_metadata.version = TEST_ROOT_VERSION - 1
    signed_root_rollback = test_bundle.generate_prod_signed_root_metadata(
        root_rollback
    )
    root_rollback_bundle = test_bundle.generate_prod_signed_bundle(
        None, signed_root_rollback
    )

    # Generates a bundle with a bad target signature.
    bad_targets_siganture = test_bundle.generate_prod_signed_bundle(
        None, prod_signed_root
    )
    # Compromises the signature.
    bad_targets_siganture.targets_metadata['targets'].signatures[0].sig = (
        b'1' * 64
    )

    # Generates a bundle with rollback attempt
    targets_rollback = test_bundle.generate_targets_metadata()
    targets_rollback.common_metadata.version = TEST_TARGETS_VERSION - 1
    targets_rollback_bundle = test_bundle.generate_prod_signed_bundle(
        targets_rollback, prod_signed_root
    )

    # Generate bundles with mismatched hash
    mismatched_hash_targets_bundles = []
    # Generate bundles with mismatched file length
    mismatched_length_targets_bundles = []
    # Generate bundles with missing hash
    missing_hash_targets_bundles = []
    # Generate bundles with personalized out payload
    personalized_out_bundles = []
    # For each of the two files in `TARGET_FILES`, we generate a number of
    # bundles each of which modify the target in the following way
    # respectively:
    # 1. Compromise its sha256 hash value in the targets metadata, so as to
    #    test hash verification logic.
    # 2. Remove the hashes, to trigger verification failure cause by missing
    #    hashes.
    # 3. Compromise the file length in the targets metadata.
    # 4. Remove the payload to emulate being personalized out, so as to test
    #    that it does not cause verification failure.
    for idx, payload_file in enumerate(TARGET_FILES.items()):
        mismatched_hash_targets = test_bundle.generate_targets_metadata()
        mismatched_hash_targets.target_files[idx].hashes[0].hash = b'0' * 32
        mismatched_hash_targets_bundle = (
            test_bundle.generate_prod_signed_bundle(
                mismatched_hash_targets, prod_signed_root
            )
        )
        mismatched_hash_targets_bundles.append(mismatched_hash_targets_bundle)

        mismatched_length_targets = test_bundle.generate_targets_metadata()
        mismatched_length_targets.target_files[idx].length = 1
        mismatched_length_targets_bundle = (
            test_bundle.generate_prod_signed_bundle(
                mismatched_length_targets, prod_signed_root
            )
        )
        mismatched_length_targets_bundles.append(
            mismatched_length_targets_bundle
        )

        missing_hash_targets = test_bundle.generate_targets_metadata()
        missing_hash_targets.target_files[idx].hashes.pop()
        missing_hash_targets_bundle = test_bundle.generate_prod_signed_bundle(
            missing_hash_targets, prod_signed_root
        )
        missing_hash_targets_bundles.append(missing_hash_targets_bundle)

        file_name, _ = payload_file
        personalized_out_bundle = test_bundle.generate_prod_signed_bundle(
            None, prod_signed_root
        )
        personalized_out_bundle.target_payloads.pop(file_name)
        personalized_out_bundles.append(personalized_out_bundle)

    with open(args.output_header, 'w') as header:
        header.write(HEADER)
        header.write(
            proto_array_declaration(dev_signed_bundle, 'kTestDevBundle')
        )
        header.write(
            proto_array_declaration(
                dev_signed_bundle_with_root, 'kTestDevBundleWithRoot'
            )
        )
        header.write(
            proto_array_declaration(
                unsigned_bundle_with_root, 'kTestUnsignedBundleWithRoot'
            )
        )
        header.write(
            proto_array_declaration(
                dev_signed_bundle_with_prod_root, 'kTestDevBundleWithProdRoot'
            )
        )
        header.write(
            proto_array_declaration(manifest_proto, 'kTestBundleManifest')
        )
        header.write(proto_array_declaration(dev_signed_root, 'kDevSignedRoot'))
        header.write(
            proto_array_declaration(prod_signed_bundle, 'kTestProdBundle')
        )
        header.write(
            proto_array_declaration(
                mismatched_root_key_and_signature_bundle,
                'kTestMismatchedRootKeyAndSignature',
            )
        )
        header.write(
            proto_array_declaration(
                signed_bad_prod_signature_bundle, 'kTestBadProdSignature'
            )
        )
        header.write(
            proto_array_declaration(
                bad_targets_siganture, 'kTestBadTargetsSignature'
            )
        )
        header.write(
            proto_array_declaration(
                targets_rollback_bundle, 'kTestTargetsRollback'
            )
        )
        header.write(
            proto_array_declaration(root_rollback_bundle, 'kTestRootRollback')
        )

        for idx, mismatched_hash_bundle in enumerate(
            mismatched_hash_targets_bundles
        ):
            header.write(
                proto_array_declaration(
                    mismatched_hash_bundle,
                    f'kTestBundleMismatchedTargetHashFile{idx}',
                )
            )

        for idx, missing_hash_bundle in enumerate(missing_hash_targets_bundles):
            header.write(
                proto_array_declaration(
                    missing_hash_bundle,
                    f'kTestBundleMissingTargetHashFile{idx}',
                )
            )

        for idx, mismatched_length_bundle in enumerate(
            mismatched_length_targets_bundles
        ):
            header.write(
                proto_array_declaration(
                    mismatched_length_bundle,
                    f'kTestBundleMismatchedTargetLengthFile{idx}',
                )
            )

        for idx, personalized_out_bundle in enumerate(personalized_out_bundles):
            header.write(
                proto_array_declaration(
                    personalized_out_bundle,
                    f'kTestBundlePersonalizedOutFile{idx}',
                )
            )
    subprocess.run(
        [
            'clang-format',
            '-i',
            args.output_header,
        ],
        check=True,
    )
    return 0


# TODO(b/237580538): Refactor the code so that each test bundle generation
# is done in a separate function or script.
# pylint: enable=too-many-locals


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools to analyze Cortex-M CPU state context captured during an exception."""

from typing import Optional, Tuple

from pw_cpu_exception_cortex_m import cortex_m_constants
from pw_cpu_exception_cortex_m_protos import cpu_state_pb2
import pw_symbolizer

# These registers are symbolized when dumped.
_SYMBOLIZED_REGISTERS = (
    'pc',
    'lr',
    'bfar',
    'mmfar',
    'msp',
    'psp',
    'r0',
    'r1',
    'r2',
    'r3',
    'r4',
    'r5',
    'r6',
    'r7',
    'r8',
    'r9',
    'r10',
    'r11',
    'r12',
)


class CortexMExceptionAnalyzer:
    """This class provides helper functions to dump a ArmV7mCpuState proto."""

    def __init__(
        self, cpu_state, symbolizer: Optional[pw_symbolizer.Symbolizer] = None
    ):
        self._cpu_state = cpu_state
        self._symbolizer = symbolizer
        self._active_cfsr_fields: Optional[
            Tuple[cortex_m_constants.BitField, ...]
        ] = None

    def active_cfsr_fields(self) -> Tuple[cortex_m_constants.BitField, ...]:
        """Returns a list of BitFields for each active CFSR flag."""

        if self._active_cfsr_fields is not None:
            return self._active_cfsr_fields

        temp_field_list = []
        if self._cpu_state.HasField('cfsr'):
            for bit_field in cortex_m_constants.PW_CORTEX_M_CFSR_BIT_FIELDS:
                if self._cpu_state.cfsr & bit_field.bit_mask:
                    temp_field_list.append(bit_field)
        self._active_cfsr_fields = tuple(temp_field_list)
        return self._active_cfsr_fields

    def is_fault_active(self) -> bool:
        """Returns true if the current CPU state indicates a fault is active."""
        if self._cpu_state.HasField('cfsr') and self._cpu_state.cfsr != 0:
            return True
        if self._cpu_state.HasField('icsr'):
            exception_number = (
                self._cpu_state.icsr
                & cortex_m_constants.PW_CORTEX_M_ICSR_VECTACTIVE_MASK
            )
            if (
                cortex_m_constants.PW_CORTEX_M_HARD_FAULT_ISR_NUM
                <= exception_number
                <= cortex_m_constants.PW_CORTEX_M_USAGE_FAULT_ISR_NUM
            ):
                return True
        return False

    def is_nested_fault(self) -> bool:
        """Returns true if the current CPU state indicates a nested fault."""
        if not self.is_fault_active():
            return False
        if (
            self._cpu_state.HasField('hfsr')
            and self._cpu_state.hfsr
            & cortex_m_constants.PW_CORTEX_M_HFSR_FORCED_MASK
        ):
            return True
        return False

    def exception_cause(self, show_active_cfsr_fields=True) -> str:
        """Analyzes CPU state to tries and classify the exception.

        Examples:
            show_active_cfsr_fields=False
              unknown exception
              memory management fault at 0x00000000
              usage fault, imprecise bus fault

            show_active_cfsr_fields=True
              usage fault [DIVBYZERO]
              memory management fault at 0x00000000 [DACCVIOL] [MMARVALID]
        """
        cause = ''
        # The CFSR can accumulate multiple exceptions.
        split_major_cause = lambda cause: cause if not cause else cause + ', '

        if self._cpu_state.HasField('cfsr') and self.is_fault_active():
            if (
                self._cpu_state.cfsr
                & cortex_m_constants.PW_CORTEX_M_CFSR_USAGE_FAULT_MASK
            ):
                cause += 'usage fault'

            if (
                self._cpu_state.cfsr
                & cortex_m_constants.PW_CORTEX_M_CFSR_MEM_FAULT_MASK
            ):
                cause = split_major_cause(cause)
                cause += 'memory management fault'
                if (
                    self._cpu_state.cfsr
                    & cortex_m_constants.PW_CORTEX_M_CFSR_MMARVALID_MASK
                ):
                    addr = (
                        '???'
                        if not self._cpu_state.HasField('mmfar')
                        else f'0x{self._cpu_state.mmfar:08x}'
                    )
                    cause += f' at {addr}'

            if (
                self._cpu_state.cfsr
                & cortex_m_constants.PW_CORTEX_M_CFSR_BUS_FAULT_MASK
            ):
                cause = split_major_cause(cause)
                if (
                    self._cpu_state.cfsr
                    & cortex_m_constants.PW_CORTEX_M_CFSR_IMPRECISERR_MASK
                ):
                    cause += 'imprecise '
                cause += 'bus fault'
                if (
                    self._cpu_state.cfsr
                    & cortex_m_constants.PW_CORTEX_M_CFSR_BFARVALID_MASK
                ):
                    addr = (
                        '???'
                        if not self._cpu_state.HasField('bfar')
                        else f'0x{self._cpu_state.bfar:08x}'
                    )
                    cause += f' at {addr}'
            if show_active_cfsr_fields:
                for field in self.active_cfsr_fields():
                    cause += f' [{field.name}]'

        return cause if cause else 'unknown exception'

    def dump_registers(self) -> str:
        """Dumps all captured CPU registers as a multi-line string."""
        registers = []
        for field in self._cpu_state.DESCRIPTOR.fields:
            if self._cpu_state.HasField(field.name):
                register_value = getattr(self._cpu_state, field.name)
                register_str = f'{field.name:<10} 0x{register_value:08x}'
                if (
                    self._symbolizer is not None
                    and field.name in _SYMBOLIZED_REGISTERS
                ):
                    symbol = self._symbolizer.symbolize(register_value)
                    if symbol.name:
                        register_str += f' {symbol}'
                registers.append(register_str)
        return '\n'.join(registers)

    def dump_active_active_cfsr_fields(self) -> str:
        """Dumps CFSR flags with their descriptions as a multi-line string."""
        fields = []
        for field in self.active_cfsr_fields():
            fields.append(f'{field.name:<11} {field.description}')
            if isinstance(field.long_description, tuple):
                long_desc = '    {}'.format(
                    '\n    '.join(field.long_description)
                )
                fields.append(long_desc)
        return '\n'.join(fields)

    def __str__(self):
        dump = [f'Exception caused by a {self.exception_cause(False)}.', '']
        if self.active_cfsr_fields():
            dump.extend(
                (
                    'Active Crash Fault Status Register (CFSR) fields:',
                    self.dump_active_active_cfsr_fields(),
                    '',
                )
            )
        else:
            dump.extend(
                (
                    'No active Crash Fault Status Register (CFSR) fields.',
                    '',
                )
            )
        dump.extend(
            (
                'All registers:',
                self.dump_registers(),
            )
        )
        return '\n'.join(dump)


def process_snapshot(
    serialized_snapshot: bytes,
    symbolizer: Optional[pw_symbolizer.Symbolizer] = None,
) -> str:
    """Returns the stringified result of a SnapshotCpuStateOverlay message run
    though a CortexMExceptionAnalyzer.
    """
    snapshot = cpu_state_pb2.SnapshotCpuStateOverlay()
    snapshot.ParseFromString(serialized_snapshot)

    if snapshot.HasField('armv7m_cpu_state'):
        state_analyzer = CortexMExceptionAnalyzer(
            snapshot.armv7m_cpu_state, symbolizer
        )
        return f'{state_analyzer}\n'

    return ''
